/**
 * Expressive TypeScript Logger for Node.js
 * @packageDocumentation
 */

/// <reference types="node" />
import { InspectOptions } from "util";

/**
 * Code frame of an error
 * @public
 * */
export declare interface ICodeFrame {
  firstLineNumber: number;
  lineNumber: number;
  columnNumber: number | null;
  linesBefore: string[];
  relevantLine: string;
  linesAfter: string[];
}

/**
 * Object representing an error with a stack trace
 * @public
 */
export declare interface IErrorObject {
  /** Is this object an error? */
  isError: true;
  /** Name of the error*/
  name: string;
  /** Error message */
  message: string;
  /** additional Error details */
  details: object;
  /** native Error object */
  nativeError: Error;
  /** Stack trace of the error */
  stack: IStackFrame[];
  /** Code frame of the error */
  codeFrame?: ICodeFrame;
}

/**
 * ErrorObject that can safely be "JSON.stringifed". All circular structures have been "util.inspected" into strings
 * @public
 */
export declare interface IErrorObjectStringifiable extends IErrorObject {
  nativeError: never;
  errorString: string;
}

/**
 * Possible style settings of utils._inspectAndHideSensitive.styles
 * Official Node.js typedefs are missing this interface.
 * @public
 */
export declare interface IHighlightStyles {
  name?: TUtilsInspectColors;
  special?: TUtilsInspectColors;
  number?: TUtilsInspectColors;
  bigint?: TUtilsInspectColors;
  boolean?: TUtilsInspectColors;
  undefined?: TUtilsInspectColors;
  null?: TUtilsInspectColors;
  string?: TUtilsInspectColors;
  symbol?: TUtilsInspectColors;
  date?: TUtilsInspectColors;
  regexp?: TUtilsInspectColors;
  module?: TUtilsInspectColors;
}

/**
 * All possible log levels
 * @public
 */
export declare interface ILogLevel {
  0: "silly";
  1: "trace";
  2: "debug";
  3: "info";
  4: "warn";
  5: "error";
  6: "fatal";
}

/**
 * All relevant information about a log message.
 * @public
 */
export declare interface ILogObject extends IStackFrame {
  /**  Optional name of the instance this application is running on. */
  instanceName?: string;
  /**  Optional name of the logger or empty string. */
  loggerName?: string;
  /** Name of the host */
  hostname: string;
  /** Optional unique request ID */
  requestId?: string;
  /**  Timestamp */
  date: Date;
  /**  Log level name (e.g. debug) */
  logLevel: TLogLevelName;
  /**  Log level ID (e.g. 3) */
  logLevelId: TLogLevelId;
  /**  Log arguments */
  argumentsArray: (IErrorObject | unknown)[];
  /**  Optional Log stack trace */
  stack?: IStackFrame[];
  /** Convert a LogObject into a JSON, is used by JSON.stringify() */
  toJSON: () => ILogObjectStringifiable;
}

/**
 * LogObject that can safely be "JSON.stringifed". All circular structures have been "util.inspected" into strings
 * @public
 */
export declare interface ILogObjectStringifiable extends ILogObject {
  argumentsArray: (IErrorObjectStringifiable | string)[];
}

/**
 * The actual settings object
 * Based on ISettingsParam, however pre-filled with defaults in case no value was provided.
 * @public
 */
export declare interface ISettings extends ISettingsParam {
  type: "json" | "pretty" | "hidden";
  instanceName?: string;
  setCallerAsLoggerName: boolean;
  name?: string;
  requestId?: string | TRequestIdFunction;
  minLevel: TLogLevelName;
  exposeStack: boolean;
  exposeErrorCodeFrame: boolean;
  exposeErrorCodeFrameLinesBeforeAndAfter: number;
  suppressStdOutput: boolean;
  overwriteConsole: boolean;
  logLevelsColors: TLogLevelColor;
  prettyInspectHighlightStyles: IHighlightStyles;
  prettyInspectOptions: InspectOptions;
  jsonInspectOptions: InspectOptions;
  dateTimePattern: string;
  dateTimeTimezone: string;
  prefix: unknown[];
  maskValuesOfKeys: (number | string)[];
  maskAny: (string | number)[];
  maskPlaceholder: string;
  printLogMessageInNewLine: boolean;
  displayDateTime: boolean;
  displayLogLevel: boolean;
  displayInstanceName: boolean;
  displayLoggerName?: boolean;
  displayRequestId: boolean;
  displayFilePath: "hidden" | "displayAll" | "hideNodeModulesOnly";
  displayFunctionName: boolean;
  displayTypes: boolean;
  stdOut: IStd;
  stdErr: IStd;
}

/**
 * Constructor: logger settings
 * all values are optional and will be pre-filled with default values
 * @public
 */
export declare interface ISettingsParam {
  /** Print log `pretty`, as a stringified `json` or prevent printing with `hidden`, default: `pretty` */
  type?: "json" | "pretty" | "hidden";
  /** Name of the instance name, default: _host name_ */
  instanceName?: string;
  /** Use the name of the caller type as the name of the logger, default: `false` */
  setCallerAsLoggerName?: boolean;
  /** Optional name of the logger instance, default: `undefined` */
  name?: string;
  /** Minimum output log level (e.g. debug), default: "silly" */
  minLevel?: TLogLevelName;
  /** Optional request ID for grouping logs by a logical request. default: `undefined` */
  requestId?: string | TRequestIdFunction;
  /** Expose stack with EVERY log message, default: `false`  */
  exposeStack?: boolean;
  /** Get Code Frame of an Error and expose it, default: `true` */
  exposeErrorCodeFrame?: boolean;
  /** Capture lines before and after a code frame, default: `5` */
  exposeErrorCodeFrameLinesBeforeAndAfter?: number;
  /** Suppress any log output to std out / std err */
  suppressStdOutput?: boolean;
  /** Catch logs going to console (e.g. console.log). Last instantiated Log instance wins */
  overwriteConsole?: boolean;
  /**  Overwrite colors of log messages of different log levels */
  logLevelsColors?: TLogLevelColor;
  /**  Overwrite colors json highlighting */
  prettyInspectHighlightStyles?: IHighlightStyles;
  /**  Options to be set for utils._inspectAndHideSensitive when output is set to pretty, default: `setting` */
  prettyInspectOptions?: InspectOptions;
  /**  Options to be set for utils._inspectAndHideSensitive when output is set to json (\{ type: "json" \}) */
  jsonInspectOptions?: InspectOptions;
  /**  DateTime pattern based on Intl.DateTimeFormat.formatToParts with additional milliseconds, default: `year-month-day hour:minute:second.millisecond` */
  dateTimePattern?: string;
  /** DateTime timezone, e.g. `utc`, or `Europe/Berlin`, `Europe/Moscow`. You can use `Intl.DateTimeFormat().resolvedOptions().timeZone` for local timezone, default: "utc" */
  dateTimeTimezone?: string;
  /**  Prefix every log message of this logger. */
  prefix?: unknown[];
  /** Exclude case-insensitive keys for object passed to `tslog` that could potentially contain sensitive information (e.g. `password` or `Authorization`), default: ["password"] */
  maskValuesOfKeys?: (number | string)[];
  /** Mask all occurrences (case-sensitive) from logs (e.g. all secrets from ENVs etc.). Will be replaced with [***] */
  maskAny?: (string | number)[];
  /** String to use a placeholder to mask sensitive values. */
  maskPlaceholder?: string;
  /** Print log message in a new line below meta information, default: `false` */
  printLogMessageInNewLine?: boolean;
  /** Display date time at the beginning of a log message, default: `true` */
  displayDateTime?: boolean;
  /** Display log level, default: `true` */
  displayLogLevel?: boolean;
  /** Display instanceName or not, default: `false` */
  displayInstanceName?: boolean;
  /** Display name of the logger. Will only be visible if `name` was set, default: `true` */
  displayLoggerName?: boolean;
  /** Display requestId or not, default: `false` */
  displayRequestId?: boolean;
  /** Display file path, default "hideNodeModulesOnly" */
  displayFilePath?: "hidden" | "displayAll" | "hideNodeModulesOnly";
  /** Display function name, default: `true`*/
  displayFunctionName?: boolean;
  /** Display type information for each attribute passed. */
  displayTypes?: boolean;
  /**  Overwrite default std out */
  stdOut?: IStd;
  /**  Overwrite default std err */
  stdErr?: IStd;
}

/**
 * All relevant information about a log message
 * @public
 */
export declare interface IStackFrame {
  /** Relative path based on the main folder */
  filePath: string;
  /** Full path */
  fullFilePath: string;
  /** Name of the file */
  fileName: string;
  /** Line number */
  lineNumber: number | null;
  /** Column Name */
  columnNumber: number | null;
  /** Called from constructor */
  isConstructor: boolean | null;
  /** Name of the function */
  functionName: string | null;
  /** Name of the class */
  typeName: string | null;
  /** Name of the Method */
  methodName: string | null;
}

/**
 * StdOut and StdErr have to implement a write function (e.g. Stream)
 * @public
 */
export declare interface IStd {
  /** stream.Writable */
  write: (message: string) => void;
}

/**
 * 📝 Expressive TypeScript Logger for Node.js
 * @public
 */
export declare class Logger {
  private readonly _logLevels;
  private _ignoreStackLevels;
  private _attachedTransports;
  private readonly _minLevelToStdErr;
  private _parentOrDefaultSettings;
  private _mySettings;
  private _childLogger;
  private _maskValuesOfKeysRegExp;
  private _maskAnyRegExp;
  /**
   * @param settings - Configuration of the logger instance  (all settings are optional with sane defaults)
   */
  constructor(settings?: ISettingsParam, parentSettings?: ISettings);
  /** Readonly settings of the current logger instance. Used for testing. */
  get settings(): ISettings;
  /**
   *  Change settings during runtime
   *  Changes will be propagated to potential child loggers
   *
   * @param settings - Settings to overwrite with. Only this settings will be overwritten, rest will remain the same.
   * @param parentSettings - INTERNAL USE: Is called by a parent logger to propagate new settings.
   */
  setSettings(settings: ISettingsParam, parentSettings?: ISettings): ISettings;
  /**
   *  Returns a child logger based on the current instance with inherited settings
   *
   * @param settings - Overwrite settings inherited from parent logger
   */
  getChildLogger(settings?: ISettingsParam): Logger;
  /**
   *  Attaches external Loggers, e.g. external log services, file system, database
   *
   * @param transportLogger - External logger to be attached. Must implement all log methods.
   * @param minLevel        - Minimum log level to be forwarded to this attached transport logger. (e.g. debug)
   */
  attachTransport(
    transportLogger: TTransportLogger<(message: ILogObject) => void>,
    minLevel?: TLogLevelName
  ): void;
  /**
   * Logs a silly message.
   * @param args  - Multiple log attributes that should be logged out.
   */
  silly(...args: unknown[]): ILogObject;
  /**
   * Logs a trace message.
   * @param args  - Multiple log attributes that should be logged out.
   */
  trace(...args: unknown[]): ILogObject;
  /**
   * Logs a debug message.
   * @param args  - Multiple log attributes that should be logged out.
   */
  debug(...args: unknown[]): ILogObject;
  /**
   * Logs an info message.
   * @param args  - Multiple log attributes that should be logged out.
   */
  info(...args: unknown[]): ILogObject;
  /**
   * Logs a warn message.
   * @param args  - Multiple log attributes that should be logged out.
   */
  warn(...args: unknown[]): ILogObject;
  /**
   * Logs an error message.
   * @param args  - Multiple log attributes that should be logged out.
   */
  error(...args: unknown[]): ILogObject;
  /**
   * Logs a fatal message.
   * @param args  - Multiple log attributes that should be logged out.
   */
  fatal(...args: unknown[]): ILogObject;
  /**
   * Helper: Pretty print error without logging it
   * @param error - Error object
   * @param print - Print the error or return only? (default: true)
   * @param exposeErrorCodeFrame  - Should the code frame be exposed? (default: true)
   * @param exposeStackTrace  - Should the stack trace be exposed? (default: true)
   * @param stackOffset - Offset lines of the stack trace (default: 0)
   * @param stackLimit  - Limit number of lines of the stack trace (default: Infinity)
   * @param std - Which std should the output be printed to? (default: stdErr)
   */
  prettyError(
    error: Error,
    print?: boolean,
    exposeErrorCodeFrame?: boolean,
    exposeStackTrace?: boolean,
    stackOffset?: number,
    stackLimit?: number,
    std?: IStd
  ): IErrorObject;
  private _handleLog;
  private _buildLogObject;
  private _buildErrorObject;
  private _toStackObjectArray;
  private _printPrettyLog;
  private _printPrettyError;
  private _printPrettyStack;
  private _printPrettyCodeFrame;
  private _logObjectToJson;
  private _printJsonLog;
  private _inspectAndHideSensitive;
  private _formatAndHideSesitive;
}

/**
 * Hex colors of different log levels
 * @public
 */
export declare type TLogLevelColor = {
  [key in TLogLevelId]: TUtilsInspectColors;
};

/**
 * Log level IDs (0 - 6)
 * @public
 */
export declare type TLogLevelId = keyof ILogLevel;

/**
 * Log level names (sill - fatal)
 * @public
 */
export declare type TLogLevelName = ILogLevel[TLogLevelId];

/**
 * Function that generates a request ID.
 * @public
 */
export declare type TRequestIdFunction = () => string;

/**
 * List of attached transport logger with their respective min log level.
 * @public
 */
export declare type TTransportLogger<T> = {
  [key in TLogLevelName]: T;
};

/**
 * Style and color options for utils._inspectAndHideSensitive.style
 * @public
 */
export declare type TUtilsInspectColors =
  | "reset"
  | "bold"
  | "dim"
  | "italic"
  | "underline"
  | "blink"
  | "inverse"
  | "hidden"
  | "strikethrough"
  | "doubleunderline"
  | "black"
  | "red"
  | "green"
  | "yellow"
  | "blue"
  | "magenta"
  | "cyan"
  | "white"
  | "bgBlack"
  | "bgRed"
  | "bgGreen"
  | "bgYellow"
  | "bgBlue"
  | "bgMagenta"
  | "bgCyan"
  | "bgWhite"
  | "framed"
  | "overlined"
  | "gray"
  | "grey"
  | "redBright"
  | "greenBright"
  | "yellowBright"
  | "blueBright"
  | "magentaBright"
  | "cyanBright"
  | "whiteBright"
  | "bgGray"
  | "bgRedBright"
  | "bgGreenBright"
  | "bgYellowBright"
  | "bgBlueBright"
  | "bgMagentaBright"
  | "bgCyanBright"
  | "bgWhiteBright";

export {};
